import os, time, math, gzip, tempfile
from collections import defaultdict
from witch_msa.configs import Configs, tqdm_styles
from witch_msa.helpers.alignment_tools import ExtendedAlignment, Alignment, read_fasta

#from multiprocessing import Lock, Pool
import concurrent.futures
from functools import partial
from tqdm import tqdm

'''
A class holding directory/alignment information for each HMM subset from the
UPP decomposition
'''
class HMMSubset(object):
    def __init__(self, path, index):
        self.alignment_dir = path
        self.index = index

        # get hmm build results from target directory
        cmd = "find {} -maxdepth 1 -name hmmbuild.input* -type f".format(path)
        #Configs.debug("Command used: {}".format(cmd))
        self.alignment_path = os.path.realpath(
                os.popen(cmd).read().split('\n')[0])
        #self.alignment = Alignment()
        #self.alignment.read_file_object(self.alignment_path)

        # also get the hmm model path
        cmd = "find {} -maxdepth 1 -name hmmbuild.model.* -type f".format(path)
        #Configs.debug("Command used: {}".format(cmd))
        self.hmm_model_path = os.path.realpath(
                os.popen(cmd).read().split('\n')[0])

        self.num_taxa = 0
        _map = {}
        with open(self.hmm_model_path, 'r') as f:
            lines = f.read().split('\n')[:15]
            for line in lines:
                key, val = line.split()[0], ''.join(line.split()[1:])
                if key == 'NSEQ':
                    self.num_taxa = int(val)
                    break

'''
Function to write an extended alignment output (for a query alignment job)
to local as intermediate outputs. These can be used as checkpoints to avoid
rerunning the jobs if the main task is interrupted.
Writing and reading to compressed format (gzip by default)
This needs to be as compact as possible, possibly only saving a few fields:
    1. the query taxon name
    2. the intermediate query alignment
'''
def writeOneCheckpointAlignment(path, query):
    # make sure we are dealing with ExtendedAlignment object
    if not isinstance(query, ExtendedAlignment):
        return
    # make sure only one taxon is present (hence, query alignment)
    if len(query) != 1:
        return
    taxon = list(query.keys())[0]; seq = query[taxon]
    line = "{}\t{}\n".format(taxon, seq)
    encoded = line.encode('utf-8')

    # every line corresponds to a query taxon and its seq, separated by a tab
    with gzip.open(path, 'ab') as f:
        f.write(encoded)

'''
Helper function to process a splice of lines (query taxa) when reading 
checkpoint alignments
'''
def readOneCheckpointAlignment(lines):
    subset_checkpoint_queries = []
    for line in lines:
        taxon = '\t'.join(line.split('\t')[:-1])
        seq = line.split('\t')[-1]
        query = ExtendedAlignment([])
        query[taxon] = seq; query._reset_col_names()
        insertion = -1; regular = 0
        for i in range(len(seq)):
            if seq[i].islower():
                query._col_labels[i] = insertion; insertion -= 1
            else:
                query._col_labels[i] = regular; regular += 1
        subset_checkpoint_queries.append(query)

    return subset_checkpoint_queries

'''
Function to read from existing checkpoint query alignments, if any exists
(e.g., ./checkpoint_alignments.txt)
Use multiple cores to simulatenously process lines
'''
def readCheckpointAlignments(path, pool, lock):
    Configs.log('Reading existing checkpoint query alignments: {}'.format(
        path))
    checkpoint_queries = {}
    s1 = time.time()

    # read from the file and decode using utf-8 (gzipped)
    raw_inputs = []
    with gzip.open(path, 'rb') as f:
        raw_inputs = f.read().decode('utf-8').split('\n')[:-1]
    num_inputs = len(raw_inputs)

    # split tasks evenly based on the number of cpus
    per_task = int(num_inputs / pool._max_workers)

    futures = []
    for i in range(0, num_inputs, per_task):
        end_ind = min(num_inputs, i + per_task)
        futures.append(pool.submit(readOneCheckpointAlignment,
            raw_inputs[i:end_ind]))
    for future in tqdm(
            concurrent.futures.as_completed(futures),
            total=num_inputs, **tqdm_styles):
        subset_checkpoint_queries = future.result()
        for query in subset_checkpoint_queries:
            checkpoint_queries[list(query.keys())[0]] = query

    runtime_read_checkpoint = time.time() - s1
    Configs.log('Finished reading {} existing checkpoint query alignments'.format(
        len(checkpoint_queries)))
    Configs.runtime(' '.join(['(readCheckpointAlignments) Time to read',
            'existing checkpoint query alignments (s):',
            str(runtime_read_checkpoint)]))
    return checkpoint_queries

'''
Function to create a local copy of the backbone alignment in upper-cases
'''
def writeTempBackbone(outdir, backbone_path):
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    tmp_backbone_path = os.path.join(outdir, 'backbone.aln.fasta')
    Configs.log('Creating a local copy of backbone alignment ' + \
            '(all letters to upper-cases) at: ' + \
            tmp_backbone_path)
    alignment = Alignment(); alignment.read_file_object(backbone_path)
    backbone_length = alignment.sequence_length()

    assert backbone_length != None, \
            'The input backbone {} is not aligned!'.format(backbone_path)

    for key in alignment.keys():
        alignment[key] = alignment[key].upper()
    alignment.write(tmp_backbone_path, 'FASTA')
    del alignment

    return tmp_backbone_path, backbone_length

'''
Function to write a single set of queries to local, given its index, etc
'''
def writeOneQuerySet(frag_names, unaligned, outdir, args):
    index, start_ind, end_ind = args
    chosen_frags = frag_names[start_ind:end_ind]
    sub_unaligned = unaligned.sub_alignment(chosen_frags)
    sub_unaligned.write(outdir + '/unaligned_frag_{}.txt'.format(index),
            'FASTA')

'''
Split and write sub-queries to local
'''
def writeSubQueries(unaligned, outdir, pool):
    #if not os.path.isdir(outdir):
    #    os.system('mkdir -p {}'.format(outdir))
    taxa_names = list(unaligned.keys())
    num_seq = len(taxa_names)

    Configs.log('Started splitting queries (N={}) to subsets...'.format(
        num_seq))
    sid_to_query_names = {}; sid_to_query_seqs = {}
    # also rename taxon name with illegal characters
    renamed_taxa = {}
    for i in range(0, num_seq):
        if '/' in taxa_names[i]:
            # replace the taxon name at the same position in taxa_names
            old_name = taxa_names[i]
            new_name = 'renamed_query_{}'.format(i)
            taxa_names[i] = new_name

            renamed_taxa[old_name] = new_name
            unaligned[new_name] = unaligned[old_name]
            unaligned.pop(old_name)

        seq = unaligned[taxa_names[i]]
        sid_to_query_names[i] = taxa_names[i]; sid_to_query_seqs[i] = seq
    Configs.log('Finished splitting queries in memory.')
    #Configs.log('Index to query map: {}'.format(str(sid_to_query_names)))
    #args = []
    #for i in range(0, num_subset):
    #    start_ind, end_ind = i * Configs.subset_size, \
    #            min(num_seq, (i + 1) * Configs.subset_size)
    #    args.append((i, start_ind, end_ind))
    #func = partial(writeOneQuerySet, frag_names, unaligned, outdir)
    #pool.map(func, args)
    #Configs.log('Finished splitting queries to local.')

    if len(renamed_taxa) > 0:
        Configs.log('The following taxa are renamed '
                '(names will be reverted in the output): '
                '{}'.format(renamed_taxa))
    return num_seq, sid_to_query_names, sid_to_query_seqs, renamed_taxa

'''
Helper function to load one alignment subset
'''
def getOneAlignmentSubset(lock, align_dir):
    this_dict = {}
    index = int(align_dir.split('/')[-1].split('_')[2])
    this_dict[index] = HMMSubset(align_dir, index) 
    lock.acquire()
    try:
        Configs.debug('Finished dealing with subset #{} from {}'.format(
            index, align_dir))
    finally:
        lock.release()

    return this_dict

'''
Load in UPP decomposition output subsets
'''
def getAlignmentSubsets(path, lock, pool):
    cmd = "find {} -name A_0_* -type d".format(path)
    #Configs.debug("Command used: {}".format(cmd))
    align_dirs = os.popen(cmd).read().split('\n')[:-1]
    Configs.log("Number of alignment subsets: {}".format(len(align_dirs)))

    # create an AlignmentSubset object for each align_dir
    index_to_hmm = {}
    func = partial(getOneAlignmentSubset, lock)
    all_dicts = list(pool.map(func, align_dirs))
    for d in all_dicts:
        index_to_hmm.update(d)
    #for align_dir in align_dirs:
    #    index = int(align_dir.split('/')[-1].split('_')[2])
    #    index_to_hmm[index] = HMMSubset(align_dir, index) 
    return index_to_hmm

'''
Read HMMSearch results given an HMMSubset object
'''
def readHMMSearch(lock, total_num_models, subset):
    lock.acquire()
    try:
        Configs.log('Reading HMMSearch result {}/{}'.format(subset.index,
            total_num_models))
    finally:
        lock.release()

    cmd = 'find {} -name hmmsearch.results.* -type f'.format(
            subset.alignment_dir)
    hmmsearch_paths = os.popen(cmd).read().split('\n')[:-1]
    subset_ranks = defaultdict(list)
    for each_path in hmmsearch_paths:
        with open(each_path, 'r') as f:
            temp_map = eval(f.read())
            for taxon, scores in temp_map.items():
                subset_ranks[taxon].append((subset.index, scores[1]))
    return subset_ranks

'''
MP version of reading in and ranking bitscores from HMM subsets (from UPP)
'''
def readAndRankBitscoreMP(index_to_hmm, renamed_taxa, lock, pool):
    ranks = defaultdict(list)
    total_num_models = len(index_to_hmm)
    Configs.log('Reading and ranking bit-scores from HMMSearch files')
    print('\nReading and ranking bit-scores from HMMSearch files...')

    # submit jobs to Pool
    func = partial(readHMMSearch, lock, total_num_models)
    all_subset_ranks, futures = [], []
    for value in index_to_hmm.values():
        futures.append(pool.submit(func, value))
    for future in tqdm(
            concurrent.futures.as_completed(futures),
            total=len(futures), **tqdm_styles):
        all_subset_ranks.append(future.result())
    #all_subset_ranks = list(pool.map(func, index_to_hmm.values()))

    # merge all MP results
    Configs.log('Ranking bit-scores')
    for subset_ranks in all_subset_ranks:
        for taxon, ind_score_pairs in subset_ranks.items():
            for pair in ind_score_pairs:
                ranks[taxon].append(pair)

    # sort the bitscores and write to local
    ranked_bitscores = defaultdict(list)
    for taxon, scores in ranks.items():
        sorted_scores = sorted(scores, key = lambda x: x[1], reverse=True)
        taxon_name = taxon
        if taxon in renamed_taxa:
            taxon_name = renamed_taxa[taxon]
        ranked_bitscores[taxon_name] = sorted_scores
    Configs.log('Finished ranking bit-scores')
    return ranked_bitscores

'''
Read in and rank bitscores from UPP decomposition
'''
#def readAndRankBitscore(index_to_hmm):
#    ranks = defaultdict(list)
#    total_num_models, counter = len(index_to_hmm), 0
#    for index, subset in index_to_hmm.items():
#        counter += 1
#        Configs.log("Reading HMMSearch result {}/{}".format(counter,
#            total_num_models))
#
#        cmd = 'find {} -name hmmsearch.results.* -type f'.format(
#            subset.alignment_dir)
#        #Configs.debug("Command used: {}".format(cmd))
#        hmmsearch_paths = os.popen(cmd).read().split('\n')[:-1]
#        for each_path in hmmsearch_paths:
#            with open(each_path, 'r') as f:
#                temp_map = eval(f.read())
#                # mapping of taxon->scores for this index
#                for taxon, scores in temp_map.items():
#                    ranks[taxon].append((index, scores[1]))
#    # write to local
#    ranked = defaultdict(list)
#    with open(Configs.outdir + '/ranked_scores.txt', 'w') as f:
#        for taxon, scores in ranks.items():
#            sorted_scores = sorted(scores, key = lambda x: x[1], reverse=True)
#            ranked[taxon] = sorted_scores
#            f.write(taxon + ':' + ';'.join(
#                [str(z) for z in sorted_scores]) + '\n')
#    Configs.warning("Finished writing ranked scores to local!")
#    return ranked

'''
Rank bit-scores
'''
def rankBitscores(index_to_hmm, renamed_taxa, lock, pool):
    s1 = time.time()
    ranked_bitscore = readAndRankBitscoreMP(index_to_hmm, renamed_taxa,
            lock, pool)
    time_rank_bitscore = time.time() - s1
    Configs.runtime(' '.join(['(rankBitscores) Time to rank bit-scores (s)',
        str(time_rank_bitscore)]))
    return ranked_bitscore

'''
Split query sequences into batches of a defined size
'''
def loadSubQueries(lock, pool):
    s1 = time.time()
    unaligned = Alignment()
    unaligned.read_file_object(Configs.query_path)
    # make sure unaligned sequences are read in upper-case letters
    for key in unaligned.keys():
        unaligned[key] = unaligned[key].upper()

    # 1) get all sub-queries, write to [outdir]/data
    data_dir = Configs.outdir + '/data'
    num_seq, sid_to_query_names, sid_to_query_seqs, \
            renamed_taxa = writeSubQueries(unaligned, data_dir, pool)
 
    # 1.2) read in all HMMSearch results (from UPP)
    index_to_hmm = getAlignmentSubsets(Configs.hmmdir, lock, pool)

    # 1.3) read and rank bitscores
    #ranked_bitscore = readAndRankBitscoreMP(index_to_hmm, renamed_taxa,
    #                                        lock, pool)

    time_load_files = time.time() - s1
    Configs.runtime(' '.join(['(loadSubQueries) Time to split queries',
            '(s):', str(time_load_files)]))
    return num_seq, index_to_hmm, sid_to_query_names, \
            sid_to_query_seqs, renamed_taxa



'''
Created on Jan 6, 2021

@author: Vlad
'''

import os
import shutil
import heapq
from configuration import Configs
from helpers import sequenceutils
from tasks import task

'''
This is where a trace is converted into a final alignment.
If necessary, the trace is compressed first.
'''

def writeAlignment(context):
    context.graph.addSingletonClusters()
    if Configs.constrain:
        compressAlignment(context)
        writeUnpackedAlignment(context)
    else:
        writeUnconstrainedAlignment(context) 

def compressAlignment(context):
    graph = context.graph
    numCells = len(context.unalignedSequences) * len(graph.clusters)
    Configs.log("Uncompressed alignment will have {} cells..".format(numCells))
    if numCells > Configs.alignmentSizeLimit*1e9:
        Configs.log("Alignment will be more than {} Gigs, compressing..".format(Configs.alignmentSizeLimit))
        compressions, numLetters = buildCompressions(context)
        startCols, startHoms = len(graph.clusters), countHomologies(graph.clusters, numLetters, set())
                
        compressedClusters, insertions = compressClusters(context, graph.clusters, compressions, numLetters, Configs.alignmentSizeLimit*1e9)
        
        endCols, endHoms = len(compressedClusters), countHomologies(compressedClusters, numLetters, insertions)
        Configs.log("Uncompressed alignment has {} columns and {} homologous pairs..".format(startCols, startHoms))
        Configs.log("Compressed alignment has {} columns and {} homologous pairs..".format(endCols, endHoms))
        Configs.log("{}% columns and {}% homologous pairs lost..".format(100*(startCols - endCols)/startCols, 100*(startHoms - endHoms)/startHoms))
        numCells = len(context.unalignedSequences) * len(compressedClusters)
        Configs.log("Compressed alignment will have {} cells..".format(numCells))
        
        graph.clusters = compressedClusters
        graph.insertions = insertions 
    
def buildInducedSubalignment(**kwargs):
    alignmentColumnsPath = kwargs["alignmentColumnsPath"]
    subalignmentPath = kwargs["subalignmentPath"]
    inducedAlignPath = kwargs["outputFile"]
    tempInducedAlignPath = os.path.join(os.path.dirname(inducedAlignPath), "temp_{}".format(os.path.basename(inducedAlignPath)))
    
    alignColumns = []
    with open(alignmentColumnsPath) as f:
        insertIdxs = set([int(token) for token in f.readline().strip().split()])
        for line in f:
            tokens = set([int(token) for token in line.strip().split()])
            alignColumns.append(tokens) 
    
    subsetAlign = sequenceutils.readFromFasta(subalignmentPath, removeDashes=False)
    inducedAlign = {taxon : ['-'] * len(alignColumns) for taxon in subsetAlign}
        
    for idx, column in enumerate(alignColumns):
        for taxon in subsetAlign:
            for c in column:
                letter = subsetAlign[taxon].seq[c]
                if letter != '-':
                    letter = letter.lower() if c in insertIdxs else letter
                    assert inducedAlign[taxon][idx] == '-'
                    inducedAlign[taxon][idx] = letter

    for s in inducedAlign:
        inducedAlign[s] = sequenceutils.Sequence(s, "".join(inducedAlign[s]))
    sequenceutils.writeFasta(inducedAlign, tempInducedAlignPath)
    shutil.move(tempInducedAlignPath, inducedAlignPath)
    
def writeUnpackedAlignment(context):
    graph = context.graph
    filePath = context.outputFile
    
    tempFile = os.path.join(os.path.dirname(filePath), "temp_{}".format(os.path.basename(filePath)))
    if os.path.exists(tempFile):
        os.remove(tempFile)
        
    clusterMap = {path : [[] for c in graph.clusters] for path in context.subalignmentPaths}
    for idx, cluster in enumerate(graph.clusters):
        for b in cluster:
            bsub, bpos = graph.matSubPosMap[b] 
            clusterMap[context.subalignmentPaths[bsub]][idx].append(bpos)
    
    inserts = {path : [] for path in context.subalignmentPaths}
    for b in graph.insertions:
        bsub, bpos = graph.matSubPosMap[b] 
        inserts[context.subalignmentPaths[bsub]].append(bpos)
        
    Configs.log("Assembling final alignment in {}".format(filePath))
    inducedSubalignTasks = []
    for bsub, subalignPath in enumerate(context.subalignmentPaths):        
        alignmentColumnsPath = os.path.join(context.graph.workingDir, "alignment_columns_{}".format(os.path.basename(subalignPath)))
        with open(alignmentColumnsPath, 'w') as textFile:
            textFile.write("{}\n".format(" ".join([str(c) for c in inserts[subalignPath]])))
            for cluster in clusterMap[subalignPath]:
                textFile.write("{}\n".format(" ".join([str(c) for c in cluster])))
        
        inducedAlignPath = os.path.join(graph.workingDir, "induced_{}".format(os.path.basename(subalignPath)))
        args = {"alignmentColumnsPath" : alignmentColumnsPath, "subalignmentPath" : subalignPath, "outputFile" : inducedAlignPath}
        inducedTask = task.Task(taskType = "buildInducedSubalignment", outputFile = args["outputFile"], taskArgs = args)
        inducedSubalignTasks.append(inducedTask)
        #inducedTask.submitTask()
    
    task.submitTasks(inducedSubalignTasks)            
    for inducedTask in task.asCompleted(inducedSubalignTasks):
        inducedAlign = sequenceutils.readFromFasta(inducedTask.outputFile, removeDashes=False)
        Configs.log("Appending induced alignment, {} sequences of length {}..".format(len(inducedAlign), len(next(iter(inducedAlign.values())).seq)))
        sequenceutils.writeFasta(inducedAlign, tempFile, append = True)   
        
        os.remove(inducedTask.taskArgs["alignmentColumnsPath"])
        os.remove(inducedTask.outputFile)
    shutil.move(tempFile, filePath)
    Configs.log("Wrote final alignment to {}".format(filePath))        
    #Configs.log("Wrote out {} clusters..".format(len(graph.clusters)))

def buildCompressions(context):
    graph = context.graph
    idxMap = {}
    for cluster in graph.clusters:
        for b in cluster:
            bsub, bpos = graph.matSubPosMap[b] 
            idxMap[context.subalignmentPaths[bsub], bpos] = b
    
    compressionTasks = []    
    for subalignPath in context.subalignmentPaths:        
        compressPath = os.path.join(context.graph.workingDir, "compression_{}".format(os.path.basename(subalignPath)))
        args = {"subalignmentPath" : subalignPath, "outputFile" : compressPath}
        compressionTask = task.Task(taskType = "compressSubalignment", outputFile = args["outputFile"], taskArgs = args)
        compressionTasks.append(compressionTask)
    task.submitTasks(compressionTasks)
    
    compressions = {}
    numLetters = {}
    for compressionTask in task.asCompleted(compressionTasks):
        with open(compressionTask.outputFile) as f:
            letterCounts = [int(token) for token in f.readline().strip().split()]
            for i, letters in enumerate(letterCounts):
                numLetters[idxMap[compressionTask.taskArgs["subalignmentPath"], i]] = letters
            for i, line in enumerate(f):
                comps = set([idxMap[compressionTask.taskArgs["subalignmentPath"], int(token)] for token in line.strip().split()])
                compressions[idxMap[compressionTask.taskArgs["subalignmentPath"], i]] = comps
        #os.remove(compressionTask.outputFile)     
    return compressions, numLetters

def combineCompressions(context, compressions):
    graph = context.graph
    
    mergedClusters = []
    compMap = {}
    for cluster in graph.clusters:
        newCluster = []
        for b in cluster:
            if b in compressions:
                center = compressions[b]
                if center not in compMap:
                    compMap[center] = newCluster
                    newCluster.append(b)
                else:
                    compMap[center].append(b)
            else:
                newCluster.append(b)
        if len(newCluster)>0:
            mergedClusters.append(newCluster)
            
    return mergedClusters

def countHomologies(clusters, numLetters, insertions):
    homs = 0
    for cluster in clusters:
        letters = 0
        for b in cluster:
            if b not in insertions:
                letters = letters + numLetters[b]
        if letters > 1:        
            homs = homs + letters * (letters - 1) / 2   
            
    return int(homs)

def compressClustersOld(context, clusters, compressions, numLetters, threshold):
    insertions = set()
    numCells = len(context.unalignedSequences) * len(clusters)
    heap = []
    
    subIdxMap = {}
    lettersMap = {i : 0 for i in range(len(clusters))}
    heapSet = set()
    
    backCompressions = {}
    for a, aset in compressions.items():
        for b in aset:
            backCompressions[b] = backCompressions.get(b, set())
            backCompressions[b].add(a)
    
    for idx, cluster in enumerate(clusters):
        for b in cluster:
            if b not in insertions:
                lettersMap[idx] = lettersMap[idx] + numLetters[b]
            subIdxMap[b] = idx
        
        dest = 0
        for b in cluster:
            for cb in compressions[b]:
                dest = max(dest, subIdxMap[cb] + 1)
                if dest >= idx:
                    break
        if dest < idx:
            heapq.heappush(heap, (lettersMap[idx], idx))
            heapSet.add(idx)
    
    while len(heap) > 0 and numCells > threshold:
        ltrs, idx = heapq.heappop(heap)
        heapSet.remove(idx)
        
        destMap = {}
        move = True
        for b in clusters[idx]:
            nbr = 0
            for cb in compressions[b]:
                nbr = max(nbr, subIdxMap[cb] + 1)
            while len(clusters[nbr]) == 0 and nbr < idx:
                nbr = nbr + 1
            destMap[b] = nbr
            if nbr >= idx:
                move = False
                break
        
        if move:
            numCells = numCells - len(context.unalignedSequences)
            for b in clusters[idx]:
                insertions.add(b)
                dest = destMap[b]
                clusters[dest].append(b)
                subIdxMap[b] = dest
                
                if b in backCompressions:
                    for nxt in backCompressions[b]:
                        nbr = subIdxMap[nxt]
                        if len(clusters[nbr]) > 0 and nbr not in heapSet:
                            heapq.heappush(heap, (lettersMap[nbr], nbr))
                            heapSet.add(nbr)
            clusters[idx] = []
    
    newClusters = [c for c in clusters if len(c) > 0]
    for c in newClusters:
        if len(c) == 1 and c[0] in insertions:
            insertions.remove(c[0])
    
    Configs.log("Compressed from {} clusters to {} clusters..".format(len(clusters), len(newClusters)))    
    return newClusters, insertions

def compressClusters(context, clusters, compressions, numLetters, threshold):
    clusters = [set(c) for c in clusters]
    insertions = set()
    numCells = len(context.unalignedSequences) * len(clusters)
    heap = []
    
    subIdxMap = {}
    lettersMap = {i : 0 for i in range(len(clusters))}
    heapSet = set()
    
    backCompressions = {}
    for a, aset in compressions.items():
        for b in aset:
            backCompressions[b] = backCompressions.get(b, set())
            backCompressions[b].add(a)
    
    for idx, cluster in enumerate(clusters):
        for b in cluster:
            if b not in insertions:
                lettersMap[idx] = lettersMap[idx] + numLetters[b]
            subIdxMap[b] = idx
        
        dest = 0
        for b in cluster:
            for cb in compressions[b]:
                dest = max(dest, subIdxMap[cb] + 1)
                if dest >= idx:
                    break
        if dest < idx:
            heapq.heappush(heap, (lettersMap[idx], idx))
            heapSet.add(idx)
    
    while len(heap) > 0 and numCells > threshold:
        ltrs, idx = heapq.heappop(heap)
        heapSet.remove(idx)
        
        move = True
        dest = idx - 1
        while len(clusters[dest]) == 0:
            dest = dest - 1
        #if dest < 0:
        #    continue
        
        destMap = {}    
        for b in clusters[idx]:
            stack = [(b, dest)]    
            while len(stack) > 0:
                curNode, curDest = stack.pop()
                destMap[curNode] = curDest

                for cb in compressions[curNode]:
                    if subIdxMap[cb] == curDest:
                        if cb not in insertions or curDest == 0:
                            move = False
                            break
                        else:
                            cd = curDest - 1
                            while len(clusters[cd]) == 0:
                                cd = cd - 1
                            stack.append((cb, cd))
                if not move:
                    break
        
        if move:
            numCells = numCells - len(context.unalignedSequences)
            for b in clusters[idx]:
                insertions.add(b)
                if b in backCompressions:
                    for nxt in backCompressions[b]:
                        nbr = subIdxMap[nxt]
                        if len(clusters[nbr]) > 0 and nbr not in heapSet:
                            heapq.heappush(heap, (lettersMap[nbr], nbr))
                            heapSet.add(nbr)
                
            for b, dest in destMap.items():
                clusters[subIdxMap[b]].remove(b)
                clusters[dest].add(b)
                subIdxMap[b] = dest
    
    newClusters = [list(c) for c in clusters if len(c) > 0]
    for c in newClusters:
        if len(c) == 1 and c[0] in insertions:
            insertions.remove(c[0])
    
    Configs.log("Compressed from {} clusters to {} clusters..".format(len(clusters), len(newClusters)))
    return newClusters, insertions

def compressSubalignment(**kwargs):
    subalignmentPath = kwargs["subalignmentPath"]
    outputPath = kwargs["outputFile"]
    tempOutputPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    
    subsetAlign = sequenceutils.readFromFasta(subalignmentPath, removeDashes=False)
    subsetLen = len(next(iter(subsetAlign.values())).seq)
    
    numLetters = []    
    compressions = []
    lastIdx = {s : -1 for s in subsetAlign} 
    for i in range(subsetLen):
        notGaps = []
        for s,v in subsetAlign.items():
            if v.seq[i] != '-':
                notGaps.append(s) 
        
        numLetters.append(len(notGaps))
        dest = set()
        for s in notGaps:
            if lastIdx[s] != -1:
                dest.add(lastIdx[s])
            lastIdx[s] = i
        compressions.append(dest)
            
    with open(tempOutputPath, 'w') as textFile:
        textFile.write("{}\n".format(" ".join([str(c) for c in numLetters])))
        for dest in compressions:
            textFile.write("{}\n".format(" ".join([str(c) for c in dest])))
    shutil.move(tempOutputPath, outputPath)

def writeUnconstrainedAlignment(context):
    graph = context.graph
    alignment = {}
    for taxon in context.unalignedSequences:            
        alignment[taxon] = sequenceutils.Sequence(taxon, ['-'] * len(graph.clusters))

    curIdxes = {taxon : 0 for taxon in context.unalignedSequences}
    for idx, cluster in enumerate(graph.clusters):
        for b in cluster:
            bsub, bpos = graph.matSubPosMap[b]
            taxon = context.subalignments[bsub][0]
            alignment[taxon].seq[idx] = context.unalignedSequences[taxon].seq[curIdxes[taxon]]
            curIdxes[taxon] = curIdxes[taxon] + 1

    for taxon in alignment:
        alignment[taxon].seq = "".join(alignment[taxon].seq)   
                            
    sequenceutils.writeFasta(alignment, context.outputFile)  
    Configs.log("Wrote final alignment to {}".format(context.outputFile)) 



'''
Created on Sep 22, 2018

@author: Vlad
'''


class Sequence:
    def __init__(self, tag, seq):
        self.tag = tag
        self.seq = seq
    
def readFromFasta(filePath, removeDashes = False):
    sequences = {}
    currentSequence = None

    with open(filePath) as f:
        for line in f:
            line = line.strip()
            if line.startswith('>'):                    
                tag = line[1:]
                currentSequence = Sequence(tag, "")
                sequences[tag] = currentSequence
            else :
                if(removeDashes):
                    line = line.replace("-", "")
                currentSequence.seq = currentSequence.seq + line

    print("Read " + str(len(sequences)) + " sequences from " + filePath + " ..")
    return sequences

def readFromFastaOrdered(filePath, removeDashes = False):
    sequences = []
    currentSequence = None

    with open(filePath) as f:
        for line in f:
            line = line.strip()
            if line.startswith('>'):                    
                tag = line[1:]
                currentSequence = Sequence(tag, "")
                sequences.append(currentSequence)
            else :
                if(removeDashes):
                    line = line.replace("-", "")
                currentSequence.seq = currentSequence.seq + line

    print("Read " + str(len(sequences)) + " sequences from " + filePath + " ..")
    return sequences

def readFromPhylip(filePath, removeDashes = False):
    sequences = {}    

    with open(filePath) as f:
        firstLine = f.readline().strip()
        
        for line in f:
            #print(line)
            tokens = line.split()
            if len(tokens) == 2:
                tag = tokens[0]
                seq = tokens[1]
                
                if(removeDashes):
                    seq = seq.replace("-", "")
                   
                if tag in sequences:
                    sequences[tag].seq = sequences[tag].seq + seq
                else:
                    sequences[tag] = Sequence(tag, seq)
                
    
    print("Read " + str(len(sequences)) + " sequences from " + filePath + " ..")                                
    return sequences

#reads match columns only
def readFromStockholm(filePath, includeInsertions = False):
    sequences = {}
    
    with open(filePath, 'r') as stockFile:
        for line in stockFile:
            line = line.strip()
            if line == "//":
                break
            elif line == "" or line[0] == "#":
                pass
            else:  
                key, seq = line.split()
                if key not in sequences:
                    sequences[key] = Sequence(key, "")
                    
                for c in seq:
                    #if includeInsertions or not (c == '.' or c in string.ascii_lowercase):
                    if includeInsertions or (c == c.upper() and c != '.'):
                        sequences[key].seq = sequences[key].seq + c    
    return sequences

def writeFasta(alignment, filePath, taxa = None, append = False):
        with open(filePath, 'a' if append else 'w') as textFile:
            if taxa is not None:
                for tag in taxa:
                    if tag in alignment:
                        textFile.write('>' + tag + '\n' + alignment[tag].seq + '\n')
            else:
                for tag in alignment:
                    textFile.write('>' + tag + '\n' + alignment[tag].seq + '\n')
      
                    
def writePhylip(alignment, filePath, taxa = None):
    maxChars = 0
    lines = []
    for tag in alignment:
        if taxa is None or tag in taxa:
            lines.append("{} {}\n".format(tag, alignment[tag].seq))
            maxChars = max(maxChars, len(alignment[tag].seq))
    
    with open(filePath, 'w') as textFile:
        textFile.write("{} {}\n".format(len(lines), maxChars))
        for line in lines:
            textFile.write(line)

def cleanGapColumns(filePath, cleanFile = None):
    align = readFromFasta(filePath, False)
    values = list(align.values())
    keepCols = []
    for i in range(len(values[0].seq)):
        for j in range(len(values)):
            if values[j].seq[i] != '-':
                keepCols.append(i)
                break
            
    print("Removing gap columns.. Kept {} out of {}..".format(len(keepCols), len(values[0].seq)))
    for s in values:
        s.seq = ''.join(s.seq[idx] for idx in keepCols)
    
    if cleanFile is None:
        cleanFile = filePath
        
    writeFasta(align, cleanFile)
    
def convertRnaToDna(filePath, destFile = None):
    align = readFromFasta(filePath, False)
    for taxon in align:
        align[taxon].seq = align[taxon].seq.replace('U', 'T')
    if destFile is None:
        destFile = filePath
    writeFasta(align, destFile)

def inferDataType(filePath):
    sequences = readFromFasta(filePath, removeDashes=True)
    acg, t, u, total = 0, 0, 0, 0
    for taxon in sequences:
        letters = sequences[taxon].seq.upper()
        for letter in letters:
            total = total + 1
            
            if letter in ('A', 'C', 'G', 'N'):
                acg = acg + 1
            elif letter == 'T':
                t = t + 1
            elif letter == 'U':
                u = u + 1
    
    if u == 0 and (acg + t)/total > 0.9:
        print("Found {}% ACGT-N, assuming DNA..".format(int(100*(acg + t)/total)))
        dataType = "dna"
    elif t == 0 and (acg + u)/total > 0.9:
        print("Found {}% ACGU-N, assuming RNA..".format(int(100*(acg + u)/total)))
        dataType = "rna"
    else:
        print("Assuming protein..")
        dataType = "protein"
          
    return dataType

def readSequenceLengthFromFasta(filePath):
    with open(filePath) as f:
        length = 0
        readSequence = False
        for line in f:
            line = line.strip()
            if line.startswith('>'):
                if readSequence:
                    return length
                readSequence = True
            else:
                length = length + len(line)
    if readSequence:
        return length

def countGaps(alignFile):
    counts = []
    currentSequence = ""

    with open(alignFile) as f:
        for line in f:
            line = line.strip()
            if line.startswith('>'): 
                
                if currentSequence is not None:
                    if len(counts) == 0:
                        counts = [0] * len(currentSequence)
                    for i in range(len(counts)):
                        if currentSequence[i] == '-':
                            counts[i] = counts[i] + 1
                                             
                currentSequence = ""
            else:
                currentSequence = currentSequence + line
        if currentSequence is not None:
            if len(counts) == 0:
                counts = [0] * len(currentSequence)
            for i in range(len(counts)):
                if currentSequence[i] == '-':
                    counts[i] = counts[i] + 1
    
    return counts



'''
Created on 10.28.2021 by Chengze Shen

Bitscore to weight calculation.
'''

import os
import time
import numpy as np
import concurrent.futures
from witch_msa.configs import Configs, tqdm_styles
from tqdm import tqdm

class Weights(object):
    weights = dict()
    weights_map = dict()
    ranked_bitscores = dict()
    def __init__(self):
        pass

'''
Function to read in weights from local given the taxon name
'''
def readWeights(taxon):
    infile = Configs.outdir + '/weights/w_{}.txt'.format(taxon)
    if not os.path.isfile(infile):
        return None, None
    else:
        weights, weights_map = [], [] 
        with open(infile, 'r') as f:
            line = f.read()
            taxon, raw = line.split(':')
            weights = [eval(x) for x in raw.split(';')]
            weights_map = {ind: w for (ind, w) in weights}
        return weights, weights_map

'''
Function to read in bitscores from local given the taxon name
'''
def readBitscores(taxon):
    infile = Configs.outdir + '/bitscores/b_{}.txt'.format(taxon)
    if not os.path.isfile(infile):
        return None, None
    else:
        bitscores = [] 
        with open(infile, 'r') as f:
            line = f.read()
            taxon, raw = line.split(':')
            bitscores = [eval(x) for x in raw.split(';')]
        return bitscores

'''
Function to calculate the HMM weighting, given the bitscores and sizes
of the HMMs (for a given query taxon)
inputs: ensemble of HMMs H (with their bitscores and sizes)
outputs: weights for HMMs H
'''
def calculateWeights(packed_data):
    taxon, indexes, bitscores, sizes = packed_data
    #logging.debug('working with: {}'.format(taxon))
    weights = {}
    
    assert len(indexes) == len(bitscores) == len(sizes)
    for i in range(len(bitscores)):
        score_i, size_i = bitscores[i], sizes[i]
        exponents = np.array(bitscores) - score_i \
                + np.log2(np.array(sizes) / size_i)
        denominator = np.sum(np.power(2, exponents))
        weights[indexes[i]] = 1. / denominator
    
    num_to_retain = min(Configs.num_hmms, len(weights))
    sorted_weights = sorted([(ind, w) for ind, w in weights.items()],
            key = lambda x: x[1], reverse=True)[:num_to_retain]
    return {taxon: tuple(sorted_weights)}

    ## write weights to local (only top k ones)
    #sorted_weights = [str(x) for x in sorted_weights]
    #with open(Configs.outdir + '/weights/w_{}.txt'.format(taxon), 'w') as f:
    #    f.write(taxon + ':' + ';'.join(sorted_weights) + '\n')
    #return None

'''
Function to write a single taxon with its ranked bitscore to local
'''
def writeQueryBitscores(packed_data):
    taxon, sorted_scores = packed_data
    str_sorted_scores = [str(x) for x in sorted_scores]

    with open(Configs.outdir + '/bitscores/b_{}.txt'.format(taxon), 'w') as f:
        f.write(taxon + ':' + ';'.join(str_sorted_scores) + '\n')
    return None

'''
Write bitscores to local (the same way as we write weights)
'''
def writeBitscores(ranked_bitscores, pool):
    s2 = time.time()
    Configs.warning('Starting to load bitscores...')
    #if not os.path.isdir(Configs.outdir + '/bitscores'):
    #    os.makedirs(Configs.outdir + '/bitscores')

    taxon_to_bitscores = {}
    for taxon, sorted_scores in ranked_bitscores.items():
        num_to_retain = min(Configs.num_hmms, len(sorted_scores))
        taxon_to_bitscores[taxon] = tuple(sorted_scores[:num_to_retain])

    #args = []
    #for taxon, sorted_scores in ranked_bitscores.items():
    #    args.append((taxon, sorted_scores))
    #all_score_temps = list(pool.map(writeQueryBitscores, args))

    time_write_scores = time.time() - s2
    Configs.warning('Finished loading bitscores in memory.')
    Configs.runtime(' '.join(['(writeBitscores) Time to write ranked bitscores',
            'to local (s):', str(time_write_scores)]))
    return taxon_to_bitscores

'''
Obtain and write weights to local based on bitscores
'''
def writeWeights(index_to_hmm, ranked_bitscores, pool):
    s2 = time.time()
    Configs.warning('Starting to calculate weights...')
    #pool = Pool(Configs.num_cpus)

    # - get sizes of each HMM
    all_sizes = {}
    for index, subset in index_to_hmm.items():
        all_sizes[index] = subset.num_taxa
        #all_sizes[index] = subset.alignment.get_num_taxa()

    # iterate through each query taxon
    # write to local for each taxon and its weights
    #if not os.path.isdir(Configs.outdir + '/weights'):
    #    os.makedirs(Configs.outdir + '/weights')
    weights, weights_map = {}, {}
    args = []
    for taxon, sorted_scores in ranked_bitscores.items():
        indexes = [x[0] for x in sorted_scores]
        bitscores = [x[1] for x in sorted_scores]
        sizes = [all_sizes[x] for x in indexes]
        args.append((taxon, indexes, bitscores, sizes))

        ## sequential version to calculate weights
        #this_weights_map,_ = calculateWeights(taxon, indexes, bitscores, sizes)
        #weights[taxon] = sorted([(ind, w) for ind, w in this_weights_map.items()],
        #        key = lambda x: x[1], reverse=True)
        #weights_map[taxon] = this_weights_map
    #all_taxon_to_weights = list(pool.map(calculateWeights, args,
    #    chunksize=Configs.chunksize))
    all_taxon_to_weights, futures = [], []
    for arg in args:
        futures.append(pool.submit(calculateWeights, arg))
    for future in tqdm(
            concurrent.futures.as_completed(futures),
            total=len(args), **tqdm_styles):
        res = future.result()
        if res:
            all_taxon_to_weights.append(res)

    taxon_to_weights = {}
    for item in all_taxon_to_weights:
        taxon_to_weights.update(item)

    time_obtain_weights = time.time() - s2
    Configs.warning('Finished calculating weights!')
    Configs.runtime(' '.join(['(writeWeights) Time to obtain weights',
            'given bitscores (s):', str(time_obtain_weights)]))
    return taxon_to_weights

'''
Write weights to local as [outdir]/weights.txt
'''
def writeWeightsToLocal(taxon_to_weights, path):
    Configs.log('Writing weights to {}'.format(path))
    with open(path, 'w') as f:
        for taxon, weights in taxon_to_weights.items():
            f.write('{}:{}\n'.format(taxon, weights))

'''
Function to read weights from a given weights path (e.g., ./weights.txt)
Return a dictionary of taxon to weights
'''
def readWeightsFromLocal(path):
    Configs.log('Reading weights from {}'.format(path))
    taxon_to_weights = {}
    with open(path, 'r') as f:
        line = f.readline()
        while line:
            # split by ':'
            taxon, taxon_weight = line.split(':')
            taxon_to_weights[taxon] = eval(taxon_weight)
            line = f.readline()
    return taxon_to_weights



#!/usr/bin/env python

# setup.py for PyPI
import setuptools
from setuptools import Command, find_packages
import os, sys, shutil
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import ArgumentParser, Namespace
from platform import platform

_root_dir = os.path.dirname(__file__)
_config_path = os.path.join(_root_dir, 'witch_msa', 'default.config')

def get_version(path):
    with open(_root_dir + '/' + path, 'r') as f:
        lines = f.read().splitlines()
        for line in lines:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]

def get_tools_dir(where=None):
    platform_name = platform()
    if where is None:
        if 'macos' in platform_name.lower():
            where = os.path.join("macOS")
        else:
            where = os.path.join("magus", "tools")
    path = os.path.join(os.getcwd(), "witch_msa", "tools", where)
    if not os.path.exists(path):
        raise OSError("Tool is not available on {}!".format(platform_name))
    return path

class ConfigWITCH(Command):
    description = "Configure WITCH for the current user and create main.config."
    user_options = [
            ('contained', 'c',
                ('Whether WITCH should be install to ~/.witch_msa '
                    'or in a self-contained manner.'))]

    def initopts(self):
        self.contained = None
        self.configfile = None
        self.basepath = None

    def initpath(self, name='main.config'):
        # install to WITCH directory
        if self.contained:
            self.configfile = os.path.expanduser(
                    os.path.abspath(os.path.join('.witch_msa', name)))
        else:
            self.configfile = os.path.expanduser('~/.witch_msa/{}'.format(name))
        self.basepath = os.path.dirname(self.configfile)
        # write to local for installation to system
        # will read in during runs to find the main.config file
        with open('witch_msa/home.path', 'w') as f:
            f.write(self.basepath)
            f.close()
        #return basepath, configfile

    def initialize_options(self):
        self.initopts()

    def finalize_options(self):
        self.initpath('main.config')
        print('\nCreating WITCH main.config at {} and tools at {}'.format(
            self.configfile, self.basepath))

    def get_tools_dest(self):
        return os.path.join(self.basepath, "tools")

    def copy_tool_to_lib(self, tool, _from, bits=True):
        _to = os.path.join(self.get_tools_dest(), tool)
        if not os.path.isdir(_from): 
            if os.path.exists(_to):
                os.rmove(_to)
            shutil.copy2(_from, _to)
        else:
            if os.path.exists(_to):
                shutil.rmtree(_to)
            shutil.copytree(_from, _to, False, None)
        return _to

    def run(self):
        '''
            prioritize software that exists in user's environment
        '''
        prioritize_user_software = True

        if not os.path.isdir(self.basepath):
            # avoid making nested directories just in case
            os.mkdir(self.basepath)
        #if not os.path.isdir(self.get_tools_dest()):
        #    os.mkdir(self.get_tools_dest())

        # create main.config file at self.configfile
        config_defaults = []
        cparser = configparser.ConfigParser()
        cparser.optionxform = str
        assert os.path.exists('{}'.format(_config_path)), \
                "default config file {} missing! Please redownload from Github\n".format(
                        _config_path)

        main_config_path = self.configfile
        if os.path.exists(main_config_path):
            print('Main configuration file {} exists...'.format(main_config_path))
            print('Overwriting existing configuration file...')

        print('\n')
        # initialize main config file using default config file
        default_config = configparser.ConfigParser()
        with open(_config_path, 'r') as f:
            default_config.read_file(f)
        for section in default_config.sections():
            cparser.add_section(section)
            for k, v in default_config[section].items():
                cparser.set(section, k, v)

        # if platform is linux then we just copy the default config file
        # as main.config
        platform_name = platform()
        tools_dir = get_tools_dir()
        set_sections = ['Basic', 'MAGUS']

        # copy magus directory to tools/
        magus_dir = os.path.join(os.getcwd(), "witch_msa", "tools", "magus")
        #copied_magus_dir = self.copy_tool_to_lib(
        #        'magus', os.path.join(os.getcwd(), "tools", "magus"))
        cparser.set('Basic', 'magus_path', magus_dir + '/magus.py')

        if 'macos' not in platform_name.lower():
            print('System is {}, using default config as main.config...'.format(
                platform_name))
            # use existing binaries from MAGUS subfolder (reduce redundancy of
            # duplicated binaries)
            #_magus_tools_dir = os.path.join(_root_dir, 'tools', 'magus', 'tools') 
            for _section in set_sections:
                # mafftpath
                cparser.set(_section, 'mafftpath',
                        os.path.join(magus_dir, 'tools', 'mafft', 'mafft'))
                # mclpath
                cparser.set(_section, 'mclpath',
                        os.path.join(magus_dir, 'tools', 'mcl', 'bin', 'mcl'))
                # fasttreepath
                cparser.set(_section, 'fasttreepath',
                        os.path.join(magus_dir, 'tools', 'fasttree', 'FastTreeMP'))
                # hmmer packages
                for hmmer_pkg in ['hmmsearch', 'hmmalign', 'hmmbuild']:
                    cparser.set(_section, '{}path'.format(hmmer_pkg),
                            os.path.join(magus_dir, 'tools', 'hmmer', hmmer_pkg))
        else:
            if 'x86' not in platform_name:
                print('Warning: system is not using x86 architecture.',
                        'Some softwares such as FastTreeMP need to be',
                        'self-provided. See {} [Basic] '.format(_config_path),
                        'section for more information.')
            print("System is {}, reconfiguring main.config...".format(platform_name))

            # configure MAGUS to use macOS compatible executables
            #_macOS_dir = os.path.join(_root_dir, 'tools', 'macOS')
            #binaries = os.popen('ls {}'.format(_macOS_dir)).read().split('\n')[:-1]
            binaries = os.listdir(tools_dir)
            for binary in binaries:
                path = os.path.join(tools_dir, binary)
                #path = os.path.join(_macOS_dir, binary)
                for _section in set_sections:
                    if 'FastTreeMP' in path:
                        cparser.set(_section, 'fasttreepath', path)
                                #self.copy_tool_to_lib('FastTreeMP', path))
                    else:
                        cparser.set(_section, '{}path'.format(binary), path)
                                #self.copy_tool_to_lib(binary, path))

        # binaries from the user's environment will be used in priority
        # if they exist
        if prioritize_user_software:
            print('Detecting existing software from the user\'s environment...')
            software = ['mafft', 'mcl', 
                    'hmmsearch', 'hmmalign', 'hmmbuild', 'FastTreeMP']
            print('\tDetected:\n')
            for soft in software:
                if shutil.which(soft):
                    print('\t{}: {}'.format(soft, shutil.which(soft)))
                    for _section in set_sections:
                        if soft == 'FastTreeMP':
                            cparser.set(_section, 'fasttreepath',
                                    shutil.which(soft))
                        elif soft == 'magus':
                            cparser.set('Basic', 'magus_path',
                                    shutil.which(soft))
                        else:
                            cparser.set(_section, '{}path'.format(soft),
                                    shutil.which(soft))

        with open(main_config_path, 'w') as f:
            cparser.write(f)
        print('\n(Done) main.config written to {}'.format(main_config_path))
        print('If you would like to make manual changes, please directly edit {}'.format(
            main_config_path))

#def create_main_config_path(prioritize_user_software=True):
#    platform_name = platform()
#    #prioritize_user_software = True
#
#    config_defaults = []
#    cparser = configparser.ConfigParser()
#    cparser.optionxform = str
#    assert os.path.exists('{}'.format(_config_path)), \
#            "default config file {} missing! Please redownload from Github\n".format(
#                    _config_path)
#
#    # initialize main.config location and record in home.path
#    basepath, configfile = initpath()
#    if not os.path.isdir(basepath):
#        os.makedirs(basepath)
#
#    #main_config_path = os.path.join(_root_dir, 'main.config')
#    main_config_path = configfile
#    if os.path.exists(main_config_path):
#        print('Main configuration file {} exists...'.format(main_config_path))
#        print('Overwriting existing configuration file...')
#        #ans = input('Do you wish to regenerate the file? [yes(y)/no(n)]:')
#        #if ans != 'yes' and ans != 'y':
#        #    print('Exiting WITCH configuration setup...')
#        #    exit(0)
#
#    print('\n\n')
#    # initialize main config file using default config file
#    default_config = configparser.ConfigParser()
#    with open(_config_path, 'r') as f:
#        default_config.read_file(f)
#    for section in default_config.sections():
#        cparser.add_section(section)
#        for k, v in default_config[section].items():
#            cparser.set(section, k, v)
#
#    # if platform is linux then we just copy the default config file
#    # as main.config
#    set_sections = ['Basic', 'MAGUS']
#    if 'macOS' not in platform_name:
#        print('System is {}, using default config as main.config...'.format(
#            platform_name))
#        # use existing binaries from MAGUS subfolder (reduce redundancy of
#        # duplicated binaries)
#        _magus_tools_dir = os.path.join(_root_dir, 'tools', 'magus', 'tools') 
#        for _section in set_sections:
#            # mafftpath
#            cparser.set(_section, 'mafftpath',
#                    os.path.join(_magus_tools_dir, 'mafft', 'mafft'))
#            # mclpath
#            cparser.set(_section, 'mclpath',
#                    os.path.join(_magus_tools_dir, 'mcl', 'bin', 'mcl'))
#            # fasttreepath
#            cparser.set(_section, 'fasttreepath',
#                    os.path.join(_magus_tools_dir, 'fasttree', 'FastTreeMP'))
#            # hmmer packages
#            for hmmer_pkg in ['hmmsearch', 'hmmalign', 'hmmbuild']:
#                cparser.set(_section, '{}path'.format(hmmer_pkg),
#                        os.path.join(_magus_tools_dir, 'hmmer', hmmer_pkg))
#    else:
#        if 'x86' not in platform_name:
#            print('Warning: system is not using x86 architecture.',
#                    'Some softwares such as FastTreeMP need to be',
#                    'self-provided. See {} [Basic] '.format(_config_path),
#                    'section for more information.')
#        print("System is {}, reconfiguring main.config...".format(platform_name))
#
#        # configure MAGUS to use macOS compatible executables
#        _macOS_dir = os.path.join(_root_dir, 'tools', 'macOS')
#
#        binaries = os.popen('ls {}'.format(_macOS_dir)).read().split('\n')[:-1]
#        for binary in binaries:
#            path = os.path.join(_macOS_dir, binary)
#            for _section in set_sections:
#                if 'FastTreeMP' in path:
#                    cparser.set(_section, 'fasttreepath', path)
#                else:
#                    cparser.set(_section, '{}path'.format(binary), path)
#
#    # binaries from the user's environment will be used in priority
#    # if they exist
#    if prioritize_user_software:
#        print('Detecting existing software from the user\'s environment...')
#        software = ['mafft', 'mcl', 
#                'hmmsearch', 'hmmalign', 'hmmbuild', 'FastTreeMP']
#        for soft in software:
#            print('\t{}: {}'.format(soft, shutil.which(soft)))
#            for _section in set_sections:
#                if shutil.which(soft):
#                    if soft != 'FastTreeMP':
#                        cparser.set(_section, '{}path'.format(soft),
#                                shutil.which(soft))
#                    else:
#                        cparser.set(_section, 'fasttreepath',
#                                shutil.which(soft))
#
#    with open(main_config_path, 'w') as f:
#        cparser.write(f)
#    print('\n(Done) main.config written to {}'.format(main_config_path))
#    print('If you would like to make manual changes, please directly edit {}'.format(
#        main_config_path))

#def setup(prioritize_user_software, platform=platform()): 
#def main():
    #create_main_config_path()
    #b_recreate_main = False
    #if len(sys.argv) > 1:
    #    b_recreate_main = sys.argv[1] == 'main'

setuptools.setup(
        #name="witch-msa",
        packages=["witch_msa", "witch_msa/gcmm", "witch_msa/helpers"], #find_packages(),
        package_data={"witch_msa": ["witch_msa/tools"]},
        include_package_data=True,
        cmdclass={"config": ConfigWITCH},
        #version=get_version('witch.py'),
        #description="WITCH - A Multiple Sequence Alignment Tool",
        long_description=open('README.rst', 'r').read(),
        long_description_content_type='text/x-rst',
        #author='Chengze Shen',
        #author_email='chengze5@illinois.edu',
        #url="https://github.com/c5shen/WITCH",
        #license="GPL-3.0",
        #classifiers=classifiers,
        #install_requires=requirements,
        #python_requires='>=3.7',
        scripts=['bin/witch-msa', 'witch.py'],
        #data_files=[("", "home.path")]
        )

#def main():
#    parser = ArgumentParser()
#    parser.add_argument('-c', '--configure', default=None,
#            choices=['macOS', 'linux'], required=False,
#            help='Specify which system the configuration file should resolve ' \
#                 'to. This is an optional field and by default, the software ' \
#                 'will resolve this automatically. Regardless, users can ' \
#                 'specify either macOS (x86) or linux.')
#    parser.add_argument('-p', '--prioritize-user-software', default="true",
#            choices=['true', 'false'], required=False,
#            help='Whether to prioritize using software in the user\'s ' \
#                 'environment. default: True')
#    args = parser.parse_args()
#    
#    if args.configure:
#        setup(args.prioritize_user_software == "true", args.configure) 
#    else:
#        setup(args.prioritize_user_software == "true")
#
#if __name__ == "__main__":
#    main()



'''
Created on Apr 14, 2020

@author: Vlad
'''

import os
import time

from helpers import sequenceutils

class Configs:
    
    workingDir = None
    sequencesPath = None
    subsetPaths = None
    subalignmentPaths = None
    backbonePaths = None
    guideTree = "fasttree"
    outputPath = None
    dataType = None
    
    decompositionMaxNumSubsets = 25
    decompositionMaxSubsetSize = 50
    decompositionStrategy = "pastastyle"
    decompositionSkeletonSize = 300
    #decompositionKmhIterations = 1
    
    graphBuildMethod = "mafft"
    graphBuildHmmExtend = False
    graphBuildRestrict = False
    graphClusterMethod = "mcl" 
    graphTraceMethod = "minclusters"
    graphTraceOptimize = False
    
    mafftRuns = 10
    mafftSize = 200
    mclInflationFactor = 4
    
    constrain = True
    onlyGuideTree = False
    recurse = True
    recurseGuideTree = "fasttree"
    recurseThreshold = 200 

    clustalPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/clustal/clustalo")
    mafftPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/mafft/mafft")
    mclPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/mcl/bin/mcl")
    mlrmclPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/mlrmcl/mlrmcl")
    hmmalignPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/hmmer/hmmalign")
    hmmbuildPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/hmmer/hmmbuild")
    hmmsearchPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/hmmer/hmmsearch")
    fasttreePath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/fasttree/FastTreeMP")
    raxmlPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tools/raxmlng/raxml-ng")
    
    logPath = None
    errorPath = None
    debugPath = None
    
    numCores = 1
    searchHeapLimit = 5000
    alignmentSizeLimit = 100
    
    @staticmethod
    def log(msg, path = None):
        print(msg)
        path = Configs.logPath if path is None else path
        Configs.writeMsg(msg, path)
    
    @staticmethod
    def error(msg, path = None):
        Configs.log(msg)
        path = Configs.errorPath if path is None else path
        Configs.writeMsg(msg, path)
    
    @staticmethod
    def debug(msg, path = None):
        path = Configs.debugPath if path is None else path
        Configs.writeMsg(msg, path)
    
    @staticmethod
    def writeMsg(msg, path):
        if path is not None:
            with open(path, 'a') as logFile:
                logFile.write("{}    {}\n".format(time.strftime("%Y-%m-%d %H:%M:%S"), msg))
    
    @staticmethod
    def inferDataType(sequencesFile):
        if Configs.dataType is None:
            Configs.dataType = sequenceutils.inferDataType(sequencesFile)
            Configs.log("Data type wasn't specified. Inferred data type {} from {}".format(Configs.dataType.upper(), sequencesFile))
        return Configs.dataType 

def buildConfigs(args):
    Configs.outputPath = os.path.abspath(args.output)
    
    if args.directory is not None:
        Configs.workingDir = os.path.abspath(args.directory) 
    else:
        Configs.workingDir = os.path.join(os.path.dirname(Configs.outputPath), "magus_working_dir")
    if not os.path.exists(Configs.workingDir):
        os.makedirs(Configs.workingDir)
    
    Configs.sequencesPath = os.path.abspath(args.sequences) if args.sequences is not None else Configs.sequencesPath
    
    Configs.guideTree = os.path.abspath(args.guidetree) if args.guidetree is not None else Configs.guideTree
    if args.guidetree is not None:
        Configs.guideTree = os.path.abspath(args.guidetree) if os.path.exists(os.path.abspath(args.guidetree)) else args.guidetree
    
    Configs.subalignmentPaths = []
    for p in args.subalignments:
        path = os.path.abspath(p)
        if os.path.isdir(path):
            for filename in os.listdir(path):
                Configs.subalignmentPaths.append(os.path.join(path, filename))
        else:
            Configs.subalignmentPaths.append(path)
    
    Configs.backbonePaths = []
    for p in args.backbones:
        path = os.path.abspath(p)
        if os.path.isdir(path):
            for filename in os.listdir(path):
                Configs.backbonePaths.append(os.path.join(path, filename))
        else:
            Configs.backbonePaths.append(path)

    if args.numprocs > 0:
        Configs.numCores = args.numprocs
    else:
        Configs.numCores = os.cpu_count()

    Configs.decompositionMaxSubsetSize = args.maxsubsetsize
    Configs.decompositionMaxNumSubsets = args.maxnumsubsets
    Configs.decompositionStrategy = args.decompstrategy
    Configs.decompositionSkeletonSize = args.decompskeletonsize
    Configs.dataType = args.datatype
    
    Configs.graphBuildMethod = args.graphbuildmethod
    Configs.graphBuildHmmExtend = args.graphbuildhmmextend.lower() == "true"
    Configs.graphBuildRestrict = args.graphbuildrestrict.lower() == "true"
    Configs.graphClusterMethod = args.graphclustermethod
    Configs.graphTraceMethod = args.graphtracemethod
    Configs.graphTraceOptimize = args.graphtraceoptimize.lower() == "true"

    Configs.mafftRuns = args.mafftruns
    Configs.mafftSize = args.mafftsize
    Configs.mclInflationFactor = args.inflationfactor
    
    Configs.constrain = args.constrain.lower() == "true"
    Configs.onlyGuideTree = args.onlyguidetree.lower() == "true"
    Configs.recurse = args.recurse.lower() == "true"
    Configs.recurseGuideTree = args.recurseguidetree
    Configs.recurseThreshold = args.recursethreshold 

    Configs.logPath = os.path.join(Configs.workingDir, "log.txt")    
    Configs.errorPath = os.path.join(Configs.workingDir, "log_errors.txt")
    Configs.debugPath = os.path.join(Configs.workingDir, "log_debug.txt")
    
    Configs.alignmentSizeLimit = args.alignsizelimit

    '''
        6.1.2021 - added by Chengze Shen
        new config for taking in the backbone weights
    '''
    Configs.backboneWeightsPath = args.backboneWeightsPath
    '''
        6.8.2022 - added by Chengze Shen
        new config for customized MCL/MAFFT path
    '''
    if args.mclpath:
        Configs.mclPath = os.path.abspath(args.mclpath)
    if args.mafftpath:
        Configs.mafftPath = os.path.abspath(args.mafftpath)
    if args.hmmalignpath:
        Configs.hmmalignPath = os.path.abspath(args.hmmalignpath)
    if args.hmmbuildpath:
        Configs.hmmbuildPath = os.path.abspath(args.hmmbuildpath)
    if args.hmmsearchpath:
        Configs.hmmsearchPath = os.path.abspath(args.hmmsearchpath)
    if args.fasttreepath:
        Configs.fasttreePath = os.path.abspath(args.fasttreepath)



'''
Created on 1.22.2022 by Chengze Shen

Global configuration class.
'''

import os
import time
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import ArgumentParser, Namespace
from platform import platform
from witch_msa.helpers.alignment_tools import inferDataType
from witch_msa.init_configs import init_config_file

# detect home.path location
homepath = os.path.dirname(__file__) + '/home.path'
_root_dir, main_config_path = init_config_file(homepath)

# default settings for tqdm progress bar style
tqdm_styles = {
        'desc': '\tRunning...', 'ascii': False,
        'ncols': 80, 
        #'disable': True,
        #'colour': 'green',
        'mininterval': 0.5
        }

'''
Configurations defined by users
'''
class Configs:
    global _root_dir

    hmmdir = None
    input_path = None
    backbone_path = None
    backbone_tree_path = None
    query_path = None
    outdir = None
    output_path = None

    chunksize = 1

    keeptemp = False
    keep_decomposition = True
    #keepsubalignment = False
    
    # WITCH configurations
    mode = 'witch-ng'
    num_hmms = 10
    use_weight = True 
    save_weight = False
    alignment_size = 10
    #weight_adjust = 'none'
    #subset_size = 1
    num_cpus = -1
    max_concurrent_jobs = None
    molecule = 'dna'
    #collapse_singletons = True

    # hmmalign/hmmsearch/magus paths
    magus_path = None
    #gcm_path = os.path.join(_root_dir, 'tools/gcm137/gcm137')
    mafftpath = None
    fasttreepath = None
    mclpath = None
    hmmsearchpath = None
    hmmalignpath = None
    hmmbuildpath = None

    log_path = None
    error_path = None
    debug_path = None
    runtime_path = None

    # MAGUS/GCM configurations
    keepgcmtemp = False
    inflation_factor = 4
    graphclustermethod = 'mcl'
    graphtracemethod = 'minclusters'
    graphtraceoptimize = 'false'
    timeout = 120

    @staticmethod
    def warning(msg, path=None):
        path = Configs.log_path if path is None else path
        Configs.write(msg, 'WARNING', path)

    @staticmethod
    def log(msg, path=None):
        path = Configs.log_path if path is None else path
        Configs.write(msg, 'LOG', path)

    @staticmethod
    def debug(msg, path=None):
        path = Configs.debug_path if path is None else path
        Configs.write(msg, 'DEBUG', path)

    @staticmethod
    def error(msg, path=None):
        path = Configs.error_path if path is None else path
        Configs.write(msg, 'ERROR', path)
    
    @staticmethod
    def runtime(msg, path=None):
        path = Configs.runtime_path if path is None else path
        with open(path, 'a') as f:
            f.write('{}\n'.format(msg))

    @staticmethod
    def write(msg, level, path):
        if path is not None:
            with open(path, 'a') as f:
                f.write('{}\t[{}] {}\n'.format(time.strftime('%Y-%m-%d %H:%M:%S'),
                    level, msg))

    @staticmethod
    def inferDataType(path):
        if Configs.molecule is None: 
            Configs.molecule = inferDataType(path)
            Configs.log('Molecule type was not specified. Inferred type: {}'.format(
                Configs.molecule))
        return Configs.molecule

# check for valid configurations and set them
def set_valid_configuration(name, conf):
    assert isinstance(conf, Namespace), \
            'Looking for Namespace object but find {}'.format(type(conf))

    # backbone alignment settings
    if name == 'Backbone':
        for k in conf.__dict__.keys():
            attr = getattr(conf, k)
            if not attr:
                continue

            if k == 'alignment_method':
                assert str(attr).lower() in ['magus', 'pasta', 'mafft'], \
                    'Backbone alignment method {} not implemented'.format(attr)
            elif k == 'backbone_size':
                assert int(attr) > 0, 'Backbone size needs to be > 0'
            elif k == 'selection_strategy':
                assert str(attr).lower() in ['median_length', 'random'], \
                    'Selection strategy {} not implemented'.format(attr)
            elif k == 'path':
                assert os.path.exists(os.path.realpath(str(attr))), \
                    '{} does not exist'.format(os.path.realpath(str(attr)))
        setattr(Configs, name, conf)
    # settings that change basic Configs class variables such as:
    # fasttreepath, hmmalignpath, etc.
    elif name == 'Basic':
        for k in conf.__dict__.keys():
            attr = getattr(conf, k)
            if not attr:
                continue
            # set variable [k] to [attr] if provided
            setattr(Configs, k, attr)
    elif name == 'MAGUS':
        setattr(Configs, name, conf)

# valid attribute check
def valid_attribute(k, v):
    assert isinstance(k, str)
    if isinstance(v, staticmethod):
        return False
    if not k.startswith('_'):
        return True
    return False

# print a list of all configurations
def getConfigs():
    print('\n********* Configurations **********')
    print('\thome.path: {}'.format(homepath))
    print('\tmain.config: {}\n'.format(main_config_path))
    for k, v in Configs.__dict__.items():
        if valid_attribute(k, v):
            print('\tConfigs.{}: {}'.format(k, v))

'''
Read in from config file if it exists. Any cmd-line provided configs will
override the config file.

Original functionality comes from SEPP -> sepp/config.py
'''
def _read_config_file(filename, opts, expand=None):
    Configs.debug('Reading config from {}'.format(filename))
    config_defaults = []
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    cparser.read_file(filename)

    if cparser.has_section('commandline'):
        for k, v in cparser.items('commandline'):
            config_defaults.append('--{}'.format(k))
            config_defaults.append(v)

    for section in cparser.sections():
        if section == 'commandline':
            continue
        if getattr(opts, section, None):
            section_name_space = getattr(opts, section)
        else:
            section_name_space = Namespace()
        for k, v in cparser.items(section):
            if expand and k == 'path':
                v = os.path.join(expand, v)
            section_name_space.__setattr__(k, v)
        opts.__setattr__(section, section_name_space)
    return config_defaults

'''
Build configurations
'''
#def buildConfigs(parser, cmdline_args):
def buildConfigs(args):
    ## read in from config file first
    #global main_config_path
    #opts = Namespace()
    #main_cmd_defaults = []

    #if os.path.exists(main_config_path):
    #    #print('Found main configuration file at {}, '.format(
    #    #    main_config_path) + 'loading in...')
    #    with open(main_config_path, 'r') as cfile:
    #        main_cmd_defaults = _read_config_file(cfile, opts)
    #input_args = main_cmd_defaults + cmdline_args

    #args = parser.parse_args(input_args, namespace=opts)

    if args.input_path != None:
        Configs.input_path = os.path.realpath(args.input_path)
    if args.hmmdir != None:
        Configs.hmmdir = os.path.realpath(args.hmmdir)
    if args.backbone_path != None:
        Configs.backbone_path = os.path.realpath(args.backbone_path)
    if args.backbone_tree_path != None:
        Configs.backbone_tree_path = os.path.realpath(args.backbone_tree_path)
    if args.query_path != None:
        Configs.query_path = os.path.realpath(args.query_path)
    
    Configs.outdir = os.path.realpath(args.outdir)
    if not os.path.exists(Configs.outdir):
        os.makedirs(Configs.outdir)
    Configs.output_path = os.path.join(Configs.outdir, args.output_path)

    Configs.keeptemp = args.keeptemp
    Configs.keep_decomposition = args.keep_decomposition == 1
    #Configs.keepsubalignment = args.keepsubalignment

    Configs.log_path = os.path.join(Configs.outdir, 'log.txt')
    Configs.error_path = os.path.join(Configs.outdir, 'error.txt')
    Configs.debug_path = os.path.join(Configs.outdir, 'debug.txt')
    Configs.runtime_path = os.path.join(Configs.outdir, 'runtime_breakdown.txt')

    # query alignment mode. Default WITCH uses MAGUS/GCM, while I also
    # implemented the WITCH-ng way.
    # TODO: Implement a smart way that mixes the two for runtime, since
    # running Smith-Waterman can be slow on super long sequences
    Configs.mode = args.mode

    if args.num_hmms > 0:
        Configs.num_hmms = args.num_hmms

    Configs.use_weight = args.use_weight == 1
    Configs.save_weight = args.save_weight == 1
    Configs.alignment_size = args.alignment_size
    #Configs.weight_adjust = args.weight_adjust 

    #Configs.chunksize = args.chunksize

    if args.num_cpus > 0:
        Configs.num_cpus = args.num_cpus
    else:
        Configs.num_cpus = os.cpu_count()

    if args.max_concurrent_jobs:
        Configs.max_concurrent_jobs = args.max_concurrent_jobs
    else:
        Configs.max_concurrent_jobs = max(100, 10 * Configs.num_cpus)

    Configs.molecule = args.molecule

    #Configs.collapse_singletons = args.collapse_singletons == 1

    # MAGUS/GCM options
    #Configs.keepgcmtemp = args.keepgcmtemp
    #Configs.inflation_factor = args.inflation_factor
    #Configs.graphclustermethod = args.graphclustermethod
    #Configs.graphtracemethod = args.graphtracemethod
    #Configs.graphtraceoptimize = args.graphtraceoptimize
    
    # additional MAGUS/GCM failsafe option to timeout a MAGUS process
    # if [timeout] seconds are reached before the process finished
    Configs.timeout = args.timeout

    # add any additional arguments to Configs
    for k in args.__dict__.keys():
        if k not in Configs.__dict__:
            k_attr = getattr(args, k)

            # check whether the configuration is valid
            set_valid_configuration(k, k_attr)
            #setattr(Configs, k, k_attr)

    ## backbone options
    #if args.backbone_size != None:
    #    setattr(Configs.backbone, 'backbone_size', args.backbone_size)
    #if args.selection_strategy != None:
    #    setattr(Configs.backbone, 'selection_strategy', args.selection_strategy)
    #if args.backbone_method != None:
    #    setattr(Configs.backbone, 'alignment_method', args.backbone_method)

    ## Bandaid way of logging main.config to log.txt
    #Configs.log('WITCH is running with: {}'.format(' '.join(cmdline_args)))
    #if os.path.exists(main_config_path):
    #    Configs.log('Main configuration loaded from {}'.format(
    #        main_config_path))



#!/usr/bin/env python3
import os, sys, time
from argparse import ArgumentParser, Namespace

from witch_msa.configs import _root_dir, main_config_path, _read_config_file 
from witch_msa.configs import *
from witch_msa.gcmm.gcmm import mainAlignmentProcess
from witch_msa.helpers.general_tools import SmartHelpFormatter

__version__ = "1.0.1"

def witch_runner():
    parser = _init_parser()
    cmdline_args = sys.argv[1:]
    
    global _root_dir, main_config_path 
    opts = Namespace()
    main_cmd_defaults = []

    # generate main.config using default setting if it is missing
    #if not os.path.exists(main_config_path):
    #    print('main.config not found, generating {}...'.format(main_config_path))
    #    os.system('python3 {}/setup.py main'.format(_root_dir))
    with open(main_config_path, 'r') as cfile:
        main_cmd_defaults = _read_config_file(cfile, opts)

    input_args = main_cmd_defaults + cmdline_args
    args = parser.parse_args(input_args, namespace=opts)

    buildConfigs(args)
    getConfigs()

    Configs.log('WITCH is running with: {}'.format(' '.join(cmdline_args)))
    if os.path.exists(main_config_path):
        Configs.log('Main configuration loaded from {}'.format(
            main_config_path))

    # run the codes
    s1 = time.time()
    mainAlignmentProcess(args)
    s2 = time.time()

    Configs.log('WITCH finished in {} seconds...'.format(s2 - s1))
    print('\nAll done! WITCH finished in {} seconds...'.format(s2 - s1))

def _init_parser():
    parser = ArgumentParser(
            description=(
                "This program runs WITCH, an alignment method "
                "extended from UPP and MAGUS."),
            conflict_handler='resolve',
            formatter_class=SmartHelpFormatter)
    parser.add_argument('-v', '--version', action='version',
            version="%(prog)s " + __version__)

    # basic settings
    basic_group = parser.add_argument_group(
            "Basic parameters".upper(),
            ' '.join(["These are basic fields.",
                "Users can choose to provide the backbone alignment/tree,",
                "or even the path to the eHMM from a previous UPP run.",
                "Otherwise, WITCH will generate them."]))
    parser.groups = dict()
    parser.groups['basic_group'] = basic_group
    basic_group.add_argument('-i', '--input-path', type=str,
            help='Path to the input unaligned file (all sequences).', required=False)
    basic_group.add_argument('-p', '--hmmdir', type=str,
            help='Path to the HMMs directory generated from UPP', required=False)
    basic_group.add_argument('-b', '--backbone-path', type=str,
            help='Path to the backbone alignment', required=False)
    basic_group.add_argument('-e', '--backbone-tree-path', type=str,
            help='Path to the backbone tree', required=False)
    basic_group.add_argument('-q', '--query-path', type=str,
            help='Path to the queries file that we want to align', required=False)
    basic_group.add_argument('-d', '--outdir', type=str,
            help='Output directory, default: witch_output/', required=False,
            default='witch_output')
    basic_group.add_argument('-o', '--output-path', type=str,
            help='Output file name, default: merged.fasta', required=False,
            default='merged.fasta')
    basic_group.add_argument('-t', '--num-cpus', type=int,
            help='Number of cpus for multi-processing. Default: -1 (all)',
            required=False, default=-1)
    basic_group.add_argument('--max-concurrent-jobs', type=int,
            help=' '.join(['Maximum number of concurrently running jobs.'
                    'Default: max(100, 10 * num_cpus)']),
            required=False, default=None)
    basic_group.add_argument('--timeout', type=int,
            help=' '.join(['Retry a query alignment after [timeout] seconds.',
                    'The retry will always use \'witch-ng mode\' (see WTICH',
                    'option below). Default: 120']),
            default=120, required=False)
    #basic_group.add_argument('--chunksize', type=int,
    #        help='Chunksize for multiprocessing', required=False,
    #        default=1)
    #basic_group.add_argument('--collapse-singletons', type=int,
    #        default=1, required=False,
    #        help='Collapse singleton columns and make them lower cases. default: 1')

    ## backbone alignment/tree options
    #backbone_group = parser.add_argument_group(
    #        "backbone options".upper(),
    #        ' '.join(["These options control how backbone sequences",
    #            "are selected and aligned."]))
    #parser.groups['backbone_group'] = backbone_group
    #backbone_group.add_argument('--backbone-size', type=int,
    #        help='Number of sequences to include in the backbone, ' \
    #                + 'default: min(1000, len(all_taxa))', default=None)
    #backbone_group.add_argument('--selection-strategy', type=str,
    #        choices=['median_length', 'random'],
    #        help='Backbone sequence selection strategy, default: ' \
    #                + 'sequences with lengths 25% around the median',
    #                default=None)
    #backbone_group.add_argument('--backbone-method', type=str,
    #        choices=['magus', 'pasta', 'mafft'],
    #        help='Alignment method on the backbone sequences, default: ' \
    #                + 'magus', default=None)

    # WITCH options
    witch_group = parser.add_argument_group(
            "WITCH options".upper(),
            ' '.join(["These options are used to customize WITCH",
                "pipeline."]))
    parser.groups['witch_group'] = witch_group
    witch_group.add_argument('-m', '--mode', type=str,
            help=' '.join(['Mode for aligning query sequences.',
                    '\nIn the case of adding query sequences into a large',
                    'backbone alignment, the original script used by WITCH',
                    'is inefficient and can lead to extremely long runtime.',
                    '\nThe re-implementation of WITCH-ng\'s way of adding',
                    'sequences (i.e., with DP algorithm) now enables WITCH',
                    'to be much faster and memory-efficient (with some',
                    'additional tweaks). It is advised to use',
                    '\'witch-ng mode\' (enabled as default) within WITCH.',
                    '\nDefault: witch-ng']),
            required=False, default='witch-ng',
            choices=['old-witch', 'witch-ng'])
    witch_group.add_argument('--keeptemp', action='store_const', const=True,
            help='Keep ALL temporary files in the process (constraints' \
                    + ', backbones, HMMSearch results, GCM results, etc.)',
            default=False)
    witch_group.add_argument('--keep-decomposition', type=int,
            help=' '.join(['Whether to keep the tree decomposition',
                    '(including temp backbone/frag files and',
                    'HMMBuild/HMMSearch results). Default: 1']),
            default=1, required=False)
    #witch_group.add_argument('--keepsubalignment',
    #        action='store_const', const=True,
    #        help='Keep all subalignments by MAGUS/GCM', default=False)
    witch_group.add_argument('-k', '--num-hmms', type=int,
            help=' '.join(['The number of top-scored HMMs used for aligning',
                    'a query. Default: 10']),
            required=False, default=10)
    witch_group.add_argument('-w', '--use-weight',
            type=int, required=False,
            help='Whether to use adjusted bitscore (weights). Default: 1',
            default=1)
    witch_group.add_argument('--save-weight',
            type=int, required=False,
            help='Whether to save weights to [outdir]/weights.txt. Default: 0',
            default=0)
    witch_group.add_argument('-A', '--alignment-size', type=int,
            help=' '.join(['Minimum decomposition subset size for tree',
                    'decomposition. Default: 10 (as UPP)']),
            default=10, required=False)
    witch_group.add_argument('--molecule', type=str,
            help='Whether input is amino/dna/rna. Default: infer from data',
            required=False, default=None, choices=['amino', 'dna', 'rna'])
    #witch_group.add_argument('-s', '--subset-size', type=int,
    #        help='Number of queries in a single GCM run, default: 1',
    #        required=False, default=1)
    #witch_group.add_argument('--weight-adjust', type=str, required=False,
    #        default='none', choices=['none', 'normalize', 'maxto1'],
    #        help='(DEPRECATED) Optional adjustment of weights, default: none')

    # GCM option
    #gcm_group = parser.add_argument_group(
    #        'MAGUS/GCM options'.upper(),
    #        ' '.join(['These options are used to customize MAGUS/GCM,',
    #            'for example the graph trace method.']))
    #parser.groups['gcm_group'] = gcm_group
    #gcm_group.add_argument('--keepgcmtemp', action='store_const',
    #        const=True, default=False, required=False,
    #        help='Keep temporary files generated from MAGUS/GCM')
    #gcm_group.add_argument('-f', '--inflation-factor', type=float,
    #        default=4, required=False,
    #        help="Inflation factor for MCL.\n[default: 4]") 
    #gcm_group.add_argument('--graphclustermethod', type=str,
    #        choices=['mcl', 'mlrmcl', 'rg', 'none'],
    #        default='mcl', required=False,
    #        help="Method for initial clustering of the alignment graph.\n[default: mcl]")
    #gcm_group.add_argument('--graphtracemethod', type=str,
    #        choices=['minclusters', 'mwtgreedy', 'mwtsearch', 'fm', 'rg', 'rgfast'],
    #        default='minclusters', required=False,
    #        help="Method for finding a trace from the alignment graph.\n[default: minclusters]")
    #gcm_group.add_argument('--graphtraceoptimize', type=str,
    #        choices=['true', 'false'], required=False, default='false',
    #        help="Run an optimization step on the graph trace.\n[default: false]")
    return parser



import re, os, subprocess, psutil, shutil
import time
import math
from collections import defaultdict
from witch_msa.configs import Configs
from witch_msa.gcmm.weighting import readWeights, readBitscores
from witch_msa.helpers.alignment_tools import Alignment, ExtendedAlignment, compressInsertions
from multiprocessing import Queue#, Lock

## initialize the lock for asynchronous safe logging
## Lock is passed from the main process
#def init_lock(l):
#    global lock
#    lock = l

'''
helper function to obtain merged query alignment with insertions marked
'''
def getQueryAlignment(query_name, path, dtype='fasta'):
    query = ExtendedAlignment([])
    query_name, insertions, query_aligned_columns = \
            query.read_query_alignment(query_name, path, dtype)

    # only remove temp file if input type is a fasta file
    if dtype == 'fasta':
        os.system('rm {}'.format(path))
        query_aligned_columns = []
    return query, query_aligned_columns

'''
Helper function to generate backbone alignments for the constraint sets
'''
def getBackbones(index_to_hmm, taxon, taxon_ind, seq, query_path, sorted_weights, 
        workdir, backbone_dir, use_gcm=True):
    ret = dict()

    # for each taxon and its scores on HMMs, sort it in decreasing order
    #for taxon, sorted_weights in weights.items():
    #taxon = next(iter(unaligned)); seq = unaligned[taxon]
    weights_map = {ind: w for (ind, w) in sorted_weights}

    ret_str = ''
    if ret_str != '':
        ret_str += '\n'

    if len(sorted_weights) == 0:
        return 'N/A', None

    # during loading bit-scores/weights, we already trimmed weights/scores
    # so here we just use them
    top_k_hmms = sorted_weights
    if Configs.use_weight:
        pass
        #if Configs.weight_adjust == 'normalize':
        #    cur_total_w = sum([w[1] for w in top_k_hmms])
        #    top_k_hmms = [(w[0], w[1] * (1. / cur_total_w))
        #            for w in top_k_hmms]
        #elif Configs.weight_adjust == 'maxto1':
        #    max_w = top_k_hmms[0][1]
        #    top_k_hmms = [(w[0], w[1] / max_w) for w in top_k_hmms]
    else:
        top_k_hmms = [(w[0], 1) for w in top_k_hmms]
    ret_str += '{}\tpassed to main pipeline with top {} weights: {}'.format(
            taxon, len(top_k_hmms), top_k_hmms)
    for item in top_k_hmms:
        # denote that this HMM will be used
        ret[item[0]] = 1

    # now we can split the fragments into chunks, each chunk corresponds to
    # all fragments added to an HMM i
    # For each HMM, run HMMAlign on its fragment chunk to get a backbone
    # alignment
    weights_path = workdir + '/weights.txt'
    #weights_file = open(weights_path, 'w')
    to_write = ''

    #fn_to_weight = dict()
    
    ### 6.21.2023 - alternative getBackbone without writing to local (used in GCM)
    # return a list of aligned positions for each hmmalign extended alignment
    subset_to_aligned_columns = dict()
    
    # initialize the backbone index at backbone_start_index
    bb_index = 0
    for i in ret.keys():
        if not os.path.isdir(workdir):
            os.makedirs(workdir)

        # [NEW] save each single sequence to a fasta
        this_hmm = index_to_hmm[i].hmm_model_path
    
        # hmmalign
        hmmalign_result_path = '{}/hmmalign.results.{}.{}.out'.format(
                workdir, taxon_ind, i)
        cmd = '{} -o {} {} {}'.format(Configs.hmmalignpath,
                hmmalign_result_path, this_hmm, query_path)
        os.system(cmd)
        #subprocess.run(cmd.split(' '))
    
        # Extended alignment
        ap_aln = ExtendedAlignment([taxon])

        # if subsequently using GCM/MAGUS to merge query-HMM alignments
        # need to save the files to local with weights.txt
        if use_gcm:
            ap_aln.build_extended_alignment(index_to_hmm[i].alignment_path,
                    [hmmalign_result_path], True)
            for key in ap_aln.keys():
                ap_aln[key] = ap_aln[key].upper()
            # save extended alignment to backbone directory
            this_bb_path = backbone_dir + '/{}_{}.fasta'.format(taxon_ind, i)
            ap_aln.write(this_bb_path, 'FASTA')

            # add the weights of the bb to weights.txt
            if Configs.use_weight:
                real_this_bb_path = os.path.realpath(this_bb_path)
                to_write += '{},{}\n'.format(real_this_bb_path,
                    weights_map[i])
                #weights_file.write('{},{}\n'.format(
                #        real_this_bb_path, weights_map[i]))
                #fn_to_weight[real_this_bb_path] = weights_map[i]
        else:
            ap_aln.read_extended_alignment(hmmalign_result_path)
            query_aligned_columns = []
            regular = 0
            for _i in range(len(ap_aln[taxon])):
                char = ap_aln[taxon][_i]
                if char == '-':
                    regular += 1
                    continue
                elif char.islower():
                    query_aligned_columns.append(-1)
                else:
                    query_aligned_columns.append(regular)
                    regular += 1

            #query_only_aln, _query_aligned_columns = \
            #        getQueryAlignment(taxon, ap_aln, dtype='mutablealignment')
            subset_to_aligned_columns[i] = query_aligned_columns

    if use_gcm:
        with open(weights_path, 'w') as weights_file:
            weights_file.write(to_write)
    
    return ret_str, weights_map, subset_to_aligned_columns #fn_to_weight

'''
The core function of GCM+eHMMs. Merge the HMMAlign results of the set of
queries to the input alignment.

The main differences between GCM+eHMMs and UPP are: 1) GCM+eHMMs uses adjusted
bitscore, which takes the number of queries in an HMM into consideration. 2)
GCM+eHMMs can utilize more than one HMM (while UPP uses the best HMM based on
bitscore) to align the queries; hence, more information is used.
'''
def alignSubQueries(backbone_path, backbone_length, index_to_hmm, lock, timeout,
        taxon, seq, query_weights, index):
    # index maps to query in unaligned 
    # query_weights in the form ((hmmX, scoreX), (hmmY, scoreY), ...)

    s11 = time.time()
    # add constraints
    constraints_dir = Configs.outdir + '/constraints/{}'.format(index)
    if not os.path.isdir(constraints_dir):
        os.makedirs(constraints_dir)
    # 0-th constraint set comes from the input alignment
    shutil.copyfile(backbone_path, '{}/c0.fasta'.format(
        constraints_dir))

    # only one query sequence in unaligned -> our second constraint
    unaligned = Alignment()
    unaligned[taxon] = seq
    query_path = constraints_dir + '/c1.fasta'
    #taxon = next(iter(unaligned))
    #unaligned[taxon] = unaligned[taxon].upper()
    unaligned.write(query_path, 'FASTA')

    #c_index = 1
    #for name in unaligned.keys():
    #    single_frag = unaligned.sub_alignment([name])
    #    # enforce upper case letters for all bases
    #    single_frag[name] = single_frag[name].upper()
    #    single_frag.write('{}/c{}.fasta'.format(constraints_dir, c_index),
    #            'FASTA')
    #    c_index += 1
    time_obtain_constraints = time.time() - s11

    s12 = time.time()
    # backbone alignments for GCM
    bb_dir = Configs.outdir + '/backbone_alignments/{}'.format(index)
    if not os.path.isdir(bb_dir):
        os.makedirs(bb_dir)
    # hmmsearch directory
    search_dir = Configs.outdir + '/search_results/{}'.format(index)
    if not os.path.isdir(search_dir):
        os.makedirs(search_dir)

    # get backbones with the information we have
    weights_str, weights_map, _ = getBackbones(index_to_hmm,
            taxon, index, seq, query_path,
            query_weights, search_dir, bb_dir)
    time_obtain_backbones = time.time() - s12

    #if weights_map:
    #    list_of_bb = list(weights_map.keys())
    #    list_of_weights = list(weights_map.values())
    #else:
    #    list_of_bb, list_of_weights = [], []

    s13 = time.time()

    # time out the process if exceeding [timeout] seconds
    task_timed_out = False
    gcm_outdir = Configs.outdir + '/magus_outputs/{}'.format(index)
    gcmpath = Configs.gcm_path

    est_path = None
    if weights_str != 'N/A':
        # run GCM (modified MAGUS which takes in weights) on the subset
        est_path = Configs.outdir + '/temp/magus_result_{}.txt'.format(index)
        cmd = [Configs.magus_path, '-np', '1',
                '-d', gcm_outdir, '-s', constraints_dir, '-b', bb_dir,
                '-o', est_path, '-f', str(Configs.inflation_factor),
                '--graphclustermethod', Configs.graphclustermethod,
                '--graphtracemethod', Configs.graphtracemethod,
                '--graphtraceoptimize', Configs.graphtraceoptimize]
        #import glob
        #files = glob.glob(f'{constraints_dir}/*')
        #cmd = [gcmpath, 'merge',
        #        '-i', *files, '-g', *list_of_bb,
        #        '-o', est_path, '-w', *[str(w) for w in list_of_weights]]
        #print(' '.join(cmd))
        # use macOS version mcl (version 21.257) if system is macOS
        if Configs.mclpath:
            cmd += ['--mclpath', Configs.mclpath]
        if Configs.use_weight:
            weights_path = search_dir + '/weights.txt'
            cmd += ['-w', weights_path] 
        #os.system(cmd)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        try:
            p.wait(timeout)
        except subprocess.TimeoutExpired:
            task_timed_out = True
            parent = psutil.Process(p.pid)
            for c in parent.children(recursive=True):
                c.kill()
            parent.kill()
        # additional check for called process error
        except subprocess.CalledProcessError:
            ret_code = p.returncode
            res = p.communicate()
            stderr_out = res[1]

            err_to_write = 'Encountered subprocess.CalledProcessError!' \
                + '\nerror code: {}'.format(ret_code) \
                + '\ndetailed call process output:'
            print(err_to_write)
            Configs.error(err_to_write)
            for line in res[0].decode(encoding='utf-8').split('\n'):
                print(line)
                Configs.error(line)
            exit(1)
    else:
        # no matching HMMs to align the given query sequence, directly
        # return
        Configs.warning('{} in task #{}'.format(taxon, index) \
                + ' does not have any matching HMMs, skipping...')
        return ExtendedAlignment([]), index, taxon

    # remove temp folders
    if not Configs.keeptemp:
        dirs_to_delete = [search_dir, bb_dir, constraints_dir]
        for _dir in dirs_to_delete:
            if os.path.isdir(_dir):
                os.system('rm -rf {}'.format(_dir))
    if not Configs.keepgcmtemp and os.path.isdir(gcm_outdir):
        os.system('rm -rf {}'.format(gcm_outdir))
    time_gcm = time.time() - s13
    
    if not task_timed_out:
        # read in the extended alignment of the query taxon
        s14 = time.time()
        if os.path.exists(est_path):
            query, _ = getQueryAlignment(taxon, est_path)
        else:
            # put the task back to queue
            lock.acquire()
            try:
                Configs.warning(
                        'Task #{}->{} finished within time but ' \
                        'has no output, queuing for retry...'.format(
                            index, taxon))
                alignSubQueries.q.put(index)
            finally:
                lock.release()
            return None, index, taxon

        time_merged_query = time.time() - s14

        lock.acquire()
        try:
            Configs.runtime(' '.join(['(alignSubQueries,',
                    'i={}) Time to obtain'.format(index),
                    'constraints (s):', str(time_obtain_constraints)]))
            Configs.runtime(' '.join(['(alignSubQueries,',
                    'i={}) Time to obtain backbones (s):'.format(index),
                    str(time_obtain_backbones)]))
            Configs.runtime(' '.join(['(alignSubQueries,',
                    'i={}) Time to run GCM and'.format(index),
                    'clean temporary files (s):', str(time_gcm)]))
            Configs.runtime(' '.join(['(alignSubQueries,',
                    'i={}) Time to obtain the merged alignment'.format(index),
                    ' (s):', str(time_merged_query)]))
            Configs.debug("[MAGUS] Command used: {}".format(' '.join(cmd)))
            Configs.log(weights_str)
            #Configs.log('{} passed to main pipeline...'.format(taxon))
        finally:
            lock.release()
        return query, index, taxon
    else:
        lock.acquire()
        try:
            Configs.debug("[MAGUS] Command used: {}".format(' '.join(cmd)))
            Configs.warning('Task #{}->{} failed, queuing for retry...'.format(
                index, taxon))
            alignSubQueries.q.put(index)
        finally:
            lock.release()
        return None, index, taxon

'''
New query-HMM alignment merging function that does not rely on GCM/MAGUS.
This one will use the knowledge of the index mapping between each sub-alignment
to the backbone alignment to create an alignment graph, and add weights to edges
between nodes (of backbone and a query).

Then, the merged alignment for a query is obtained by running Smith-Waterman
on the alignment graph (|C|x|q|, where |C| is the # sites of the backbone and 
|q| the length of the query sequence.

All of the above can be achieved without calls to subprocesses such as MAGUS
or the experimental GCM code (by Baqiao), and most operations can be done
in-memory (for speed).
'''
def alignSubQueriesNew(backbone_path, backbone_length, index_to_hmm, lock, timeout,
        taxon, seq, query_weights, index):
    ############## STEP 1 ###############
    # obtain backbones (edges without weights yet)
    s11 = time.time()
    # query constraint to local (for hmmsearch)
    constraints_dir = Configs.outdir + '/constraints/{}'.format(index)
    if not os.path.isdir(constraints_dir):
        os.makedirs(constraints_dir)

    unaligned = Alignment(); unaligned[taxon] = seq
    query_path = constraints_dir + '/c1.fasta'
    unaligned.write(query_path, 'FASTA')

    # hmmsearch directory
    search_dir = Configs.outdir + '/search_results/{}'.format(index)
    if not os.path.isdir(search_dir):
        os.makedirs(search_dir)
    weights_str, subset_to_weight, subset_to_aligned_columns = \
            getBackbones(index_to_hmm, taxon, index, seq, query_path,
                    query_weights, search_dir, None, use_gcm=False)
    #print(taxon, subset_to_weight[next(iter(subset_to_weight))],
    #        subset_to_aligned_columns[next(iter(subset_to_aligned_columns))])
    time_obtain_backbones = time.time() - s11

    s12 = time.time()
    query = ExtendedAlignment([])
    if weights_str != 'N/A':
        ############## STEP 2 ###############
        # added weighted edges to the alignment graph from gluing alignment
        # subset_to_retained_columns -> globally shared/used
        # subset_to_nongaps_per_column -> globally shared/used
        # subset_to_weight -> k=10 then 10 subsets and their weights
        # subset_to_aligned_columns -> aligned columns in subsets for this query

        # technically we are just creating a matrix for DP later, and adding in
        # corresponding weighted edges as they are read in
        alignment_graph = [[0. for _j in range(backbone_length + 1)]
                              for _i in range(len(seq) + 1)]
        # key -> (index of query char, index of backbone col)
        # val -> accumulated weights of the key
        combined_weights = defaultdict(float)

        subset_to_retained_columns = alignSubQueriesNew.subset_to_retained_columns
        subset_to_nongaps_per_column = alignSubQueriesNew.subset_to_nongaps_per_column
        # also record the minimum/maximum of indexes of backbone used
        # this information can be used to limit the DP search space (for speed-up)
        min_col_ind, max_col_ind = backbone_length + 1, -1

        for subset, aligned_columns in subset_to_aligned_columns.items():
            # subset -> a bit complicated to explain here how it is used 
            # indexes of aligned_columns -> row index
            for i in range(len(aligned_columns)):
                subset_col = aligned_columns[i]
                if subset_col == -1:    # skip insertion site of query
                    continue
                
                # compute the column index to add edge weight
                #print(subset, subset_col, len(subset_to_retained_columns[subset]))
                j = subset_to_retained_columns[subset][subset_col]
                
                # compute the weight to accumulated to combined_weights
                weight_to_add = subset_to_nongaps_per_column[subset][subset_col] * \
                        subset_to_weight[subset]
                combined_weights[(i, j)] += weight_to_add

                # update min/max index used for the column (backbone)
                min_col_ind = min(min_col_ind, j)
                max_col_ind = max(max_col_ind, j)
        #print(taxon, seq, combined_weights)
        #exit(0)

        ############## STEP 2.5 ###############
        # run Smith-Waterman on the alignment graph to obtain the highest weighted
        # trace (thus, alignment of the query to the backbone)
        # use the min/max col indexes to bound backtrace region
        backtrace = [[0 for _j in range(backbone_length + 1)]
                              for _i in range(len(seq) + 1)]

        for i in range(0, len(seq) + 1):
            for j in range(min_col_ind, max_col_ind + 2):
                if i == 0 or j == min_col_ind:
                    alignment_graph[i][j] = 0.
                    continue

                cur_max = 0.; cur_bt = 0
                cw = combined_weights.get((i-1, j-1), 0.)
                values = [alignment_graph[i-1][j-1] + cw,
                          alignment_graph[i-1][j],
                          alignment_graph[i][j-1]]
                for _ind, val in enumerate(values):
                    if _ind == 0 and cw <= 0:
                        cur_bt = 1
                        continue
                    if val > cur_max:
                        cur_max = val
                        cur_bt = _ind
                alignment_graph[i][j] = cur_max
                backtrace[i][j] = cur_bt

        # obtain alignment from backtrace
        # the alignment is represented by the set of characters of the query
        result = []; i, j = len(seq), max_col_ind + 1
        while i > 0 and j > min_col_ind:
            # diagonal - match
            bt = backtrace[i][j]
            if bt == 0:    # a match
                result.append(seq[i-1])
                i -= 1; j -= 1
            elif bt == 1:  # a query "insertion", use lowercase
                result.append(seq[i-1].lower())
                i -= 1
            elif bt == 2:  # a query "deletion", use a gap '-'
                result.append('-')
                j -= 1
            else:
                raise ValueError

        # fill up remaining positions (i.e., i and j both to 0)
        # only one "while" will happen
        while i > 0:
            result.append(seq[i-1].lower()); i -= 1
        while j > min_col_ind:
            result.append('-'); j -= 1

        # reverse the result to get correct order
        result = result[::-1]
        #print(taxon, result)

        # append '-' to start, end to fill up the alignment (since we use min/max
        # of backbone columns to reduce search space)
        result = ['-'] * min_col_ind + result + \
                ['-'] * (backbone_length - max_col_ind - 1)

        # compress insertions at front/end of the query alignment
        combined = compressInsertions(''.join(result))

        # return an extended alignment object with just the query
        # sequence and its updated indexes
        query[taxon] = combined; query._reset_col_names()
        insertion = -1; regular = 0
        for i in range(len(combined)):
            if combined[i].islower():
                query._col_labels[i] = insertion; insertion -= 1
            else:
                query._col_labels[i] = regular; regular += 1

    # remove temp folders
    if not Configs.keeptemp:
        dirs_to_delete = [search_dir, constraints_dir]
        for _dir in dirs_to_delete:
            if os.path.isdir(_dir):
                os.system('rm -rf {}'.format(_dir))
    time_alignment_trace = time.time() - s12

    # if query is empty, meaning there is no weight and the whole algorithm
    # did not run. Simply return the empty alignment obj
    if len(query) == 0:
        lock.acquire()
        try:
            Configs.warning('{} in task #{}'.format(taxon, index) \
                    + ' does not have any matching HMMs, ignored in final output...')
        finally:
            lock.release()
        return query, index, taxon
    
    # sanity check for finishing the query-HMM alignment merge
    if query.get_length() >= backbone_length:
        lock.acquire()
        try:
            Configs.runtime(' '.join(['(alignSubQueriesNew,',
                    'i={}) Time to obtain constraint/backbones'.format(index),
                    ' (s):', str(time_obtain_backbones)]))
            Configs.runtime(' '.join(['(alignSubQueriesNew,',
                    'i={}) Time to run alignment and clean-up'.format(index),
                    ' (s):', str(time_alignment_trace)]))
            Configs.log(weights_str)
        finally:
            lock.release()
        return query, index, taxon
    else:
        lock.acquire()
        try:
            Configs.warning('Task #{}->{} failed, ignored in final output...'.format(
                index, taxon))
        finally:
            lock.release()
        # return an empty alignment as indication of failure
        return ExtendedAlignment([]), index, taxon
